# test-packmime-delaybox.tcl
#
# Based on .../ns/tcl/ex/rbp_demo.tcl by John Heidemann
#
# Round-trip time: 12ms.

# useful constants
set CLIENT 0
set SERVER 1

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Usage
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

proc usage {} {
    puts stderr {usage: ns test-packmime-delaybox.tcl [variable value]
     defaults:
       duration       100       - length of measured simulation (seconds)
       warmup         10        - warmup interval (seconds)
       rate           1600      - number of new connections / second
       packmime-count 1         - number of PackMime clouds
       queue-method   RED       - DropTail, RED, ...
       queue-length   1000      - queue length (packets)
       trg-delay      0.001	- Trageted delay at the RED queue 
                                  (make min_thresh = 12 pkt)
       bytes	      0		- packet-AQM
       
       link-delay     2ms       - propagation delay on bottleneck link
       link-bw        100Mb     - bandwidth on bottleneck link
       access-bw      1000Mb	- access link bandwidth
       access-delay   2ms	- propagation delay of the access link
       routers        1         - use routers/delayboxes in network?
       
       ecn 	      1		- ECN enable by default
       ecn-syn	      0		- No ECN in SYN/ACK pkt
       ecn-syn-wait   0		- ECN+ with noWaiting
       minrto 	      0.2	- min rto for tcp
       rtxcur-init    3.0	- initial value of rto
       window         16        - max TCP window size (packets)
       segsize        1000      - default TCP segment size (bytes)
       loss-rate      0         - maximum packet loss rate
       run            1         - experiment run number
       debug          0         - debug level
       quiet          0         - don't output status messages?
       datadir        data      - directory to store output
       gzip           0         - gzip trace-queue
       make-trace     0         - make trace file
       http-trace     0         - trace HTTP connections
       new-seed       12345	- seed for PackMime random number generator

       rsp-size-avg   3000      - average response size in bytes
    }
    exit 1
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Command-Line Option Default Values
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

proc default_options {} {
    global opts opt_wants_arg

    set raw_opt_info {
    #
    # to add a new command-line option, insert the option name and default
    # value inside this structure
    #
    duration 100
    warmup 10
    
    # rate -- total number of connections/s
    # packmime-count -- number of PackMime instances
    rate 1600
    packmime-count 5
    
    # AQM
    queue-method RED
    queue-length 1000
    trg-delay 0.001
    bytes 0
    
    link-delay 2ms
    link-bw 100Mb
    access-bw 1000Mb
    access-delay 2ms
    routers 1

    # TCP
    tcptype Sack
    window 16
    segsize 1000
    ecn 1
    ecn-syn 0
    ecn-syn-wait 0
    minrto 0.2
    rtxcur-init 3.0

    # PackMime vars
    loss-rate 0
    rsp-size-avg 3000

    # experiment run number (changing this will change RNG seeds used)
    run 1
    new-seed 12345

    # data collection
    datadir data
    make-trace 0
    http-trace 0

    quiet 0
    debug 0
    gzip 0
    }

    # parses the default options and creates the $opts array
    while {$raw_opt_info != ""} {
        if {![regexp "^\[^\n\]*\n" $raw_opt_info line]} {
            break
        }
        regsub "^\[^\n\]*\n" $raw_opt_info {} raw_opt_info
        set line [string trim $line]
        if {[regexp "^\[ \t\]*#" $line]} {
            continue
        }
        if {$line == ""} {
            continue
        } elseif [regexp {^([^ ]+)[ ]+([^ ]+)$} $line dummy key value] {
            set opts($key) $value
            set opt_wants_arg($key) 1
        } else {
            set opt_wants_arg($key) 0
            # die "unknown stuff in raw_opt_info\n"
        }
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Process Command Line Arguments
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

proc process_args {} {
    global argc argv opts opt_wants_arg

    default_options
    for {set i 0} {$i < $argc} {incr i} {
        set key [lindex $argv $i]
        if {$key == "-?" || $key == "--help" || $key == "-help" || $key == "-h"} {
            usage
        }
        regsub {^-} $key {} key
        if {![info exists opt_wants_arg($key)]} {
            puts stderr "unknown option $key";
            usage
        }
        if {$opt_wants_arg($key)} {
            incr i
            set opts($key) [lindex $argv $i]
        } else {
            set opts($key) [expr !opts($key)]
        }
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Helper function to create duplex links
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

proc my-duplex-link {ns n1 n2 bw delay queue_method queue_length} {
    $ns duplex-link $n1 $n2 $bw $delay $queue_method
    $ns queue-limit $n1 $n2 $queue_length
    $ns queue-limit $n2 $n1 $queue_length
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# MAIN
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

proc main {} {
    global argv
    process_args
    new MySim
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Instantiate the Class
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

Class MySim

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Setup network links and nodes
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc init_network {} {
    global opts
    $self instvar ns_ c_ db_ s_ fmon_
 
    # AMIT : Change for smooth dropping at RED router 
    #Queue/RED set gentle_ true

    # AMIT : Modify for packet or byte based router
    if { $opts(bytes) == 1 } {
    	Queue/RED set bytes_ true	
    	Queue/RED set queue_in_bytes_ true
    } else {
    	Queue/RED set bytes_ false	
    	Queue/RED set queue_in_bytes_ false
    }

    # AMIT : Modify for dropping or marking at router
    Queue/RED set use_mark_p_ true
    Queue/RED set mark_p_ 2.0; # Enable marking instead of dropping unless queue is full

    Queue/RED set mean_pktsize $opts(segsize)
    Queue/RED set targetdelay_ $opts(trg-delay)

    if { $opts(ecn) == 1} {
	    Queue/RED set setbit_ true
    }
    
    #  Topology
    #
    #           1000Mb               100Mb           1000Mb
    #  client  --------  delaybox  ------ delaybox ------ server
    #  cloud     2ms                 2ms            2ms   cloud
    #
    # c_(0,packmime-1)     db_(0)          db_(1)         s_(0,packmime-1)

    # build clients
    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
    	set c_($i) [$ns_ node]
    }

    # build servers
    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
    	set s_($i) [$ns_ node]
    }

    # build delayboxes and links
    if {$opts(routers) || $opts(packmime-count) > 1} {
        set db_(0) [$ns_ DelayBox]
        set db_(1) [$ns_ DelayBox]
        my-duplex-link $ns_ $db_(0) $db_(1) $opts(link-bw) $opts(link-delay) \
            $opts(queue-method) $opts(queue-length)
	    
        for {set i 0} {$i < $opts(packmime-count)} {incr i} {
            my-duplex-link $ns_ $c_($i) $db_(0) $opts(access-bw) $opts(access-delay) \
                $opts(queue-method) $opts(queue-length)
        }
        for {set i 0} {$i < $opts(packmime-count)} {incr i} {
            my-duplex-link $ns_ $s_($i) $db_(1) $opts(access-bw) $opts(access-delay) \
                $opts(queue-method) $opts(queue-length)
        }

	# Monitor how many packets marked and dropped at congested router
	set slink_ [$ns_ link $db_(1) $db_(0)]
	set fmon_ [$ns_ makeflowmon Fid]
	$ns_ attach-fmon $slink_ $fmon_
	
    } else {
        my-duplex-link $ns_ $c_(0) $s_(0) $opts(link-bw) $opts(link-delay) \
            $opts(queue-method) $opts(queue-length)
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#  Setup PackMime Traffic
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc init_connections {} {
    global opts defaultRNG CLIENT SERVER
    $self instvar ns_ s_ c_ pm_ db_ rng_ rate_ req_size rsp_size client_delay server_rsp_delay \
        server_delay client_bw server_bw flow_arrive rate_ loss_rate

    # set TCP options
    Agent/TCP set precisionReduce_ true
    Agent/TCP/FullTcp set segsize_ $opts(segsize)
    Agent/TCP set window_ $opts(window)
    Agent/TCP set rtxcur_init_ $opts(rtxcur-init)
    Agent/TCP set minrto_ $opts(minrto)
    Agent/TCP set ecn_ $opts(ecn)
    if { $opts(ecn-syn) == 1 } {
    	Agent/TCP/FullTcp set ecn_syn_ true
    }
    #if { $opts(ecn-syn-wait) == 1 } {
    #	Agent/TCP/FullTcp set ecn_syn_wait_ true
    #}
    Agent/TCP/FullTcp set ecn_syn_wait_ $opts(ecn-syn-wait)

    
    # create PackMime(s) and assign clients and server clouds
    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
       set pm_($i) [new PackMimeHTTP]
        $pm_($i) set-client $c_($i)
        $pm_($i) set-server $s_($i)
    }

    # set rate per PackMime
    set rate_ [expr [expr $opts(rate) * 1.0] / $opts(packmime-count)]
    
    # create RNGs 
    for {set i 0} {$i < 5} {incr i} {
        set rng_($i) [new RNG]
        set this_seed [expr $opts(new-seed)*($i+1)*$opts(run)]
        $rng_($i) seed $this_seed
        puts "RNG: $i seed: $this_seed"
	# advance the seeds for different experiment runs
	# for {set j 1} {$j < $opts(run)} {incr j} {
	#	$rng_($i) next-substream
	# }
    }

    # create random variables
    # AMIT : Modify here to change file size distribution
    #set req_size [new RandomVariable/PackMimeHTTPFileSize $rate_ $CLIENT]
    #set rsp_size [new RandomVariable/PackMimeHTTPFileSize $rate_ $SERVER]
    set rsp_size [new RandomVariable/ParetoII]
    $rsp_size set avg_ $opts(rsp-size-avg)
    $rsp_size set shape_ 1.2
    set req_size [new RandomVariable/Constant]
    $req_size set val_ 100; #100bytes

    set server_delay [new RandomVariable/Uniform]
    $server_delay set min_ 0;  # 0ms
    $server_delay set max_ 100; # This makes round-trip propagation delay in [12+0 -- 12+200] + Queuing delay

    set flow_arrive [new RandomVariable/PackMimeHTTPFlowArrive $rate_]
    #set flow_arrive [new RandomVariable/Exponential]
    #$flow_arrive set avg_ [expr 1.0/$rate_]
    
    #Add server response delay 
    set server_rsp_delay [new RandomVariable/Uniform]
    #$server_rsp_delay set min_ 0;  # 0ms
    #$server_rsp_delay set max_ 1; #10ms
    
    # assign RNGs to RandomVariables
    $req_size use-rng $rng_(0)
    $rsp_size use-rng $rng_(1)
    $flow_arrive use-rng $rng_(2)
    $server_delay use-rng $rng_(3)
    #$server_rsp_delay use-rng $rng_(4)

    # setup rules for DelayBoxes
    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
       $db_(0) add-rule [$c_($i) id] [$s_($i) id] $server_delay
    }

    # set PackMime variables
    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
        $pm_($i) set-rate $rate_
        $pm_($i) set-req_size $req_size
        $pm_($i) set-rsp_size $rsp_size
        $pm_($i) set-flow_arrive $flow_arrive
        #$pm_($i) set-server_delay $server_rsp_delay
        
	$pm_($i) set-debug $opts(debug)
	$pm_($i) set-warmup $opts(warmup)

	$pm_($i) set-TCP $opts(tcptype)
        if {$opts(http-trace)} {
	    # record HTTP statistics
	    $pm_($i) set-outfile "$opts(datadir)/data-test-packmime-delaybox-$i.dat"
        }
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Schedule Traffic and Packet Tracing
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc schedule_traffic {} {
    global opts
    $self instvar ns_ pm_ db_

    for {set i 0} {$i < $opts(packmime-count)} {incr i} {
      $ns_ at 0.0 "$pm_($i) start"
      $ns_ at $opts(duration) "$pm_($i) stop"

    }

    $ns_ at $opts(warmup) "$self printpackets $opts(warmup)"

    # Uncomment the next line and comment following 2 lines for tracing the queue
    $ns_ at $opts(warmup) "$self trace"
    # set end [expr $opts(duration) + $opts(warmup)]
    # $ns_ at $end "$self finish"
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Setup Tracing
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc trace {} {
    global opts
    $self instvar ns_ db_ c_ s_ trace_file_ rate_

    Trace set show_tcphdr_ 1
    if {$opts(make-trace)} {
        if {$opts(gzip)} {
            set trace_file_ [open "|gzip > $opts(datadir)/data-test-packmime-delaybox.trq.gz" w]
        } else {
            set trace_file_ [open "$opts(datadir)/data-test-packmime-delaybox.trq" w]
        }
    }

    if {$opts(make-trace)} {
        if {$opts(routers) || $opts(packmime-count) > 1} {
      	    $ns_ trace-queue $db_(1) $db_(0) $trace_file_
      	    #$ns_ trace-queue $db_(0) $db_(1) $trace_file_
        } else {
            $ns_ trace-queue $c_(0) $s_(0) $trace_file_
            $ns_ trace-queue $s_(0) $c_(0) $trace_file_
        }
    }
   
    set end [expr $opts(duration) + $opts(warmup)]
    if {$opts(make-trace)} {
        $ns_ at $end "$ns_ flush-trace ; close $trace_file_ ; $self finish"
    } else {
        $ns_ at $end "$self finish"
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Statistics
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc printpackets {time} {
    global opts 
    $self instvar ns_ fmon_
   set packets [$fmon_ set pdepartures_]
   set bytes [$fmon_ set bdepartures_]
   set dropped [$fmon_ set pdrops_]
   set marked [$fmon_ set pmarks_]
   #set droprate [expr $dropped/$packets]
   puts "time: $time packets: $packets bytes: $bytes dropped: $dropped marked: $marked"
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Cleanup
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc finish {} {
    global opts defaultRNG
    $self instvar ns_ pm_ db_ rng_ req_size rsp_size client_delay server_rsp_delay \
        server_delay client_bw server_bw flow_arrive loss_rate trace_file_ fmon_

    # stop PackMime
    #for {set i 0} {$i < $opts(packmime-count)} {incr i} {
    #  $pm_($i) stop
    #}

    set end [expr $opts(duration) + $opts(warmup)]
    $self printpackets $end

    # delete all of the RNGs and RanVars we created
    for {set i 0} {$i < 5} {incr i} {
        delete $rng_($i)
    }
    delete $req_size
    delete $rsp_size
    delete $server_delay
    #delete $server_rsp_delay
    delete $flow_arrive

    # delete PackMime
    # Sally deleted this, because it caused a segmentation fault.
    #for {set i 0} {$i < $opts(packmime-count)} {incr i} {
    #   delete $pm_($i)
    #}

    # delete DelayBox nodes
    delete $db_(0)
    delete $db_(1)

    # delete Simulator
    delete $ns_

    # post-processing
    $self process

    exit 0
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Post-Process
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc process {} {
    global opts

    if {!$opts(quiet)} {
       puts "Done with simulation, processing..."
    }

    if {!$opts(quiet)} {
      puts stderr "DONE!"
    }
}

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Initialize Simulation
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

MySim instproc init {} {
    global opts

    $self instvar ns_

    exec mkdir -p $opts(datadir)

    # setup simulator
    remove-all-packet-headers;            # removes all packet headers
    add-packet-header Flags IP TCP;       # adds TCP/IP headers 
    					  # Flags header carries ecn related bits
    set ns_ [new Simulator];              # instantiate the simulator
    $ns_ use-scheduler Heap;              # use the Heap scheduler

    # setup network and traffic
    $self init_network
    $self init_connections
    $self schedule_traffic

    # print status
    if {!$opts(quiet)} {
    set total [expr $opts(duration) + $opts(warmup)]
    set minutes  [expr $total / 60]
    puts "run $opts(run)"
    puts "$opts(rate) connections/s"
    puts "$opts(packmime-count) PackMime(s), average response $opts(rsp-size-avg) bytes"
    puts "$minutes minutes (${total} s)  warmup $opts(warmup) s"
    puts "tcptype: $opts(tcptype)"
    puts "$opts(window) packet window, $opts(segsize) Bytes segment"
    puts "rtxcur_init_: $opts(rtxcur-init) s, minrto_ $opts(minrto)"
    puts "bottleneck bw $opts(link-bw) ps, accesslink bw $opts(access-bw) ps"
    puts "bottleneck propagation delay $opts(link-delay)"
    puts "access link propagation delay $opts(access-delay)"
    puts "queue-method: $opts(queue-method) queue-length: $opts(queue-length) pkts target-delay: $opts(trg-delay) s bytes: $opts(bytes)"
    puts "ecn: $opts(ecn) ecn-syn: $opts(ecn-syn) ecn-syn-wait: $opts(ecn-syn-wait)" 
    puts "datadir: $opts(datadir)"
    }

    # start the simulation
    $ns_ run
}


main
